-- Find science pack recipes
local packs = data.raw["tool"]
local pack_recipes = {}
local final_recipes = {}

local function is_nonzero(p)
    return ((p.amount and p.amount > 0) or (p.max_amount and p.min_amount and (p.max_amount > 0 or p.min_amount > 0))) or (p[2] and p[2] > 0)
end

for k, _ in pairs(packs) do
    pack_recipes[k] = {}
end

for _, recipe in pairs(data.raw.recipe) do
    if recipe.results then
        for _, result in ipairs(recipe.results) do
            if is_nonzero(result) then
                if (result.type == "item" and packs[result.name])
                or (result[1] and packs[result[1]]) then
                    if recipe.enabled then final_recipes[result.name or result[1]] = recipe.name end
                    pack_recipes[result.name or result[1]][recipe.name] = {prototype = recipe, technologies = {}}
                end
            end
        end
    elseif recipe.result then
        if packs[recipe.result] and (recipe.result_amount and recipe.result_amount > 0 or true) then
            if recipe.enabled then  final_recipes[recipe.result] = recipe.name end
            pack_recipes[recipe.result][recipe.name] = {prototype = recipe, technologies = {}}
        end
    end
end

-- Find technologies unlocking these recipes
for _, tech in pairs(data.raw.technology) do
    if tech.effects then
        for _, effect in ipairs(tech.effects) do
            if effect.type == "unlock-recipe" then
                for _, recipes in pairs(pack_recipes) do
                    if recipes[effect.recipe] then
                        table.insert(recipes[effect.recipe].technologies, tech)
                    end
                end
            end
        end
    end
end

-- Find science pack dependencies
local visited = {}
local dependencies = {}

local function create_dependency_set()
    local set = {}
    
    for k, _ in pairs(packs) do
        set[k] = "none"
    end

    return set
end

local function find_dependencies(tech)
    if not visited[tech.name] then
        visited[tech.name] = true
        local deps = create_dependency_set()

        for _, ingredient in ipairs(tech.unit.ingredients) do
            if is_nonzero(ingredient) then
                deps[ingredient.name or ingredient[1]] = "direct"
            end
        end

        if tech.prerequisites then
            for _, dep in ipairs(tech.prerequisites) do
                find_dependencies(data.raw.technology[dep])

                for k, v in pairs(dependencies[dep]) do
                    if v ~= "none" then
                        deps[k] = "indirect"
                    end
                end
            end
        end

        dependencies[tech.name] = deps
    end
end

for _, tech in pairs(data.raw.technology) do
    find_dependencies(tech)
end

-- Filter out self-dependent pack technologies and choose final choices
for k, pack in pairs(pack_recipes) do
    if not final_recipes[k] then
        local recipes_to_delete = {}
        for recipe_name, recipe in pairs(pack) do
            local to_delete = {}
            for i, tech in ipairs(recipe.technologies) do
                if dependencies[tech.name][k] ~= "none" then
                    to_delete[#to_delete+1] = i
                end
            end

            local has_deps = #recipe.technologies > 0

            for _, v in ipairs(to_delete) do
                table.remove(recipe.technologies, v)
            end

            if has_deps and #recipe.technologies == 0 then
                recipes_to_delete[#recipes_to_delete+1] = recipe_name
            end
        end

        for _, v in ipairs(recipes_to_delete) do
            pack[v] = nil
        end

        final_recipes[k] = ({next(pack)})[2] and ({next(pack)})[2].name or nil
    end
end

-- Fix dependencies
visited = {}

local function fix(tech)
    if not visited[tech.name] then
        visited[tech.name] = true

        for k, v in pairs(dependencies[tech.name]) do
            if v == "direct" then
                tech.prerequisites = tech.prerequisites or {}
                local t = final_recipes[k] and pack_recipes[k][final_recipes[k]].technologies[1] or nil

                if t then
                    table.insert(tech.prerequisites, t)
                end
            end
        end
    end
end

for _, tech in pairs(data.raw.technology) do
    fix(tech)
end

-- Hide unavailable technologies
for _, tech in pairs(data.raw.technology) do
    if tech.prerequisites then
        local hide = false
        for _, pre in ipairs(tech.prerequisites) do
            hide = true
            break
        end

        if hide then
            tech.enabled = false
        end
    end
end